"""F&F Fox devices RestAPI client. See www.fif.com.pl/fox."""
import aiohttp
from .const import (
    API_COMMON_DEVICE_OFF,
    API_COMMON_DEVICE_ON,
    API_COMMON_GET_DEVICE_INFO,
    API_COMMON_GET_STATE,
    API_COMMON_SET_STATE,
    API_CLIENT_CONNECTION_TIMEOUT,
    API_RESPONSE_STATUS_FAIL,
    REQUEST_CHANNEL_KEY,
    REQUEST_STATE_KEY,

)
import json
import logging
from .rest_api_responses import (
    RestApiBaseResponse,
    RestApiDeviceInfoResponse,
    RestApiDeviceStateResponse,
    RestApiError
)
import requests
from requests.compat import urljoin

class RestApiClient:
    """RestAPI client implmentation.

    Provides base communication with F&F Fox devices.
    F&F Fox device supports only HTTP GET request method.
    """

    def __init__(self, host: str, api_key: str):
        """Default construcring object. Host and api_key are required to make connection."""
        self._host = host
        self._api_key = api_key
        self._base_api_url = None
        self.__response_error_hook = None
        self.__session_timeout =  aiohttp.ClientTimeout(total=None,
            sock_connect=API_CLIENT_CONNECTION_TIMEOUT, sock_read=API_CLIENT_CONNECTION_TIMEOUT)

    def register_response_error_hook(self, response_hook):
        """Register response error hook."""
        self.__response_error_hook = response_hook

    def unregister_response_error_hook(self):
        """Unregister response error hook."""
        self.__response_error_hook = None

    def get_base_api_url(self) -> str:
        """Create base_api_url if needed and return it."""
        if self._base_api_url is not None:
            return self._base_api_url
        #Fox device working with rest api key. This key must exsist in url in below format.
        url_pattern = "http://{0}/{1}/"
        self._base_api_url = url_pattern.format(self._host, self._api_key)
        return self._base_api_url

    async def async_api_get_device_state(self, channel = None) -> RestApiDeviceStateResponse:
        """Get F&F Fox device state.

        Make HTTP GET request to device and get state.
        Keyword arguments:
        channel -- optional, get state from provided channel

        Return: RestApiDeviceStateResponse
        """
        params = {}
        if channel is not None:
            params = {REQUEST_CHANNEL_KEY: channel}
        logging.info("Making call in async_api_get_device_state().")
        response_content = await self.async_make_api_call_get(API_COMMON_GET_STATE, params)
        if response_content is None:
            return RestApiDeviceStateResponse(status=API_RESPONSE_STATUS_FAIL)
        if isinstance(response_content, RestApiError):
            return RestApiDeviceStateResponse(API_RESPONSE_STATUS_FAIL, errorObj=response_content)
        return RestApiDeviceStateResponse(**json.loads(response_content))

    async def async_api_set_device_state(self, state, channel = None) -> RestApiBaseResponse:
        """Set F&F Fox device state.

        Make HTTP GET request to device and set state
        Keyword arguments:
        state -- state to set on device can be true or false - it will be marked to on/off
        channel -- optional, set state on provided channel

        Return: RestApiBaseResponse
        """
        params = {
            REQUEST_STATE_KEY: API_COMMON_DEVICE_ON if state == True else API_COMMON_DEVICE_OFF
        }
        if channel is not None:
            params.update({REQUEST_CHANNEL_KEY: channel})
        logging.info("Making call in async_api_set_device_state() with params {0}.".format(params))
        response_content = await self.async_make_api_call_get(API_COMMON_SET_STATE, params)
        if response_content is None:
            return RestApiBaseResponse(API_RESPONSE_STATUS_FAIL)
        if isinstance(response_content, RestApiError):
            return RestApiBaseResponse(API_RESPONSE_STATUS_FAIL, errorObj=response_content)
        return RestApiBaseResponse(**json.loads(response_content))

    async def async_api_get_device_info(self) -> RestApiDeviceInfoResponse:
        """Get F&F Fox device info.

        Get device information such as firmware version, name etc.
        See RestApiDeviceInfoResponse to check what data is returnig from device.
        """
        logging.info("Making call in async_api_get_device_info().")
        response_content = await self.async_make_api_call_get(API_COMMON_GET_DEVICE_INFO)
        if response_content is None:
            return RestApiDeviceInfoResponse(status=API_RESPONSE_STATUS_FAIL)
        if isinstance(response_content, RestApiError):
            return RestApiDeviceInfoResponse(status=API_RESPONSE_STATUS_FAIL, errorObj=response_content)
        return RestApiDeviceInfoResponse(**json.loads(response_content))

    async def async_make_api_call_get(self, method: str, query_params = None):
        """Make HTTP GET request by given parameters.

        Keyword arguments:
        method -- method name to make request
        query_params - optional request parameters
        """
        if not isinstance(method, str):
            logging.warning("Wrong argument passed to method. Http method accept only string values.")
            return
        if query_params is not None and not isinstance(query_params, dict):
            logging.warning("Wrong argument passed to method. Query params must be dict.")
            return
        try:
            async with aiohttp.ClientSession(timeout = self.__session_timeout) as session:
                async with session.get(urljoin(self.get_base_api_url(), method), params=query_params) as resp:
                    response = await resp.read()
                    logging.info("Received RAW response {0}".format(response))
                    self.__invoke_response_error_hook(None) #Raise reponse_error_hook with None errors.
                    return response
        except aiohttp.ClientConnectionError as e:
            logging.error(e)
            self.__invoke_response_error_hook(e)
        except requests.exceptions.RequestException as e:
            logging.error(e)
            self.__invoke_response_error_hook(e)
        return None

    def __invoke_response_error_hook(self, e):
        """Invoke response error hook if registered."""
        if self.__response_error_hook != None:
                self.__response_error_hook(e)


"""F&F Fox RGBW device implementation."""

from __future__ import annotations
from .fox_base_device import DeviceData, FoxBaseDevice
from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import RestApiBaseResponse
from foxrestapiclient.connection.const import (
    API_RESPONSE_STATUS_FAIL,
    API_RESPONSE_STATUS_OK,
)
from .const import (
    API_RGBW_SET_COLOR_HSV,
    API_RGBW_GET_COLOR_HSV
)
import json

class FoxRGBWDevice(FoxBaseDevice):
    """Fox RGBW device."""

    def __init__(self, device_data: DeviceData):
        """Initalize object."""
        super().__init__(device_data.name, device_data.host, device_data.api_key,
                        device_data.mac_addr, device_data.type)
        self.hsv_color = [0, 0, 0]
        self.__device_api_client = self.DeviceRestApiImplementer(super())
        self._state = False

    class HSVColorData(RestApiBaseResponse):
        """HSV color data holder."""

        def __init__(self, h: str = None, s: str = None, v: str = None,
                    status: str = API_RESPONSE_STATUS_FAIL) -> None:
            """Initalize object."""
            super().__init__(status)
            self.hue = self.try_parse_to_int(h)
            self.saturation = self.try_parse_to_int(s)
            #Brightness value
            self.value = int((self.try_parse_to_int(v) * 255) / 100)

        def try_parse_to_int(self, value):
            """Try parse given value to int.

            Keyword arguments:
            value -- value to parse.
            Return: parsed value to int or 0 if parsing went wrong.
            """
            try:
                return int(value)
            except:
                return 0

    class DeviceRestApiImplementer:
        """Specific RestAPI methods definition used by device."""

        def __init__(self, restApiClient: RestApiClient) -> None:
            """Initialize object."""
            self._restApiClient = restApiClient

        async def async_get_hsv_color(self) -> FoxRGBWDevice.HSVColorData:
            """Get HSV values by given channel."""
            device_response = await self._restApiClient.async_make_api_call_get(API_RGBW_GET_COLOR_HSV)
            if device_response is None:
                return FoxRGBWDevice.HSVColorData(status=API_RESPONSE_STATUS_FAIL)
            return FoxRGBWDevice.HSVColorData(**json.loads(device_response))

        async def async_set_hsv_color(self, params = None) -> RestApiBaseResponse:
            """Set HSV values by given channel."""
            device_response = await self._restApiClient.async_make_api_call_get(API_RGBW_SET_COLOR_HSV, params)
            if device_response is None:
                return RestApiBaseResponse(status=API_RESPONSE_STATUS_FAIL)
            return RestApiBaseResponse(**json.loads(device_response))

    def get_hs_color(self):
        """Get HS values in following format [hue, saturation].

        Return: list with hue and saturation values.
        """
        hs = [self.hsv_color[0], self.hsv_color[1]]
        return hs
    def get_hsv_color(self):
        """Get HSV color.

        Return: list with hue, saturation, value.
        """
        return self.hsv_color

    def get_brightness(self):
        """Get brightness value.

        Return: brightness.
        """
        return self.hsv_color[2]

    async def async_fetch_color_hsv(self):
        """Fetch HSV color from device.

        Return: list with hue, saturation and value. If error occured 0,0,0 will be returned.
        """
        hsv_data = await self.__device_api_client.async_get_hsv_color()
        if hsv_data.status != API_RESPONSE_STATUS_OK:
            self._state = False
            return [0, 0, 0]
        self.is_available = True
        self.hsv_color = [hsv_data.hue, hsv_data.saturation, hsv_data.value]
        return self.hsv_color

    async def async_set_brightness(self, brightness):
        """Set brightness to device.

        Keyword arguments:
        brightness -- value in range <0,100>
        """
        await self.async_set_color_hsv(value=brightness)

    async def async_set_color_hsv(self, hue = None, saturation = None, value = None):
        """Set HSV color on device.

        Keyword arguments:
        hue -- value from range <0,359>
        saturation -- value from range <0,100>
        value -- value from range <0,100>
        """
        params = {}
        if hue is not None:
            params.update({"h": int(hue) if hue < 360 else 359})
        if saturation is not None:
            params.update({"s": int(saturation)})
        if value is not None:
            params.update({"v": int(100 * ((int(value)) / 255))})
        device_response = await self.__device_api_client.async_set_hsv_color(params)
        if device_response.status != API_RESPONSE_STATUS_OK:
            self._state = False
            return

    def is_on(self, channel: int = None):
        """Return device is on status."""
        return self._state

    async def async_fetch_update(self):
        """Fetch all available data from device."""
        self._state = await self.async_fetch_channel_state()
        await self.async_fetch_color_hsv()



"""F&F Fox R1S1 device implementation."""

from .fox_base_device import DeviceData, FoxBaseDevice
from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import RestApiBaseResponse
from foxrestapiclient.connection.const import (
    API_RESPONSE_STATUS_FAIL,
)
from .const import (
    API_R1S1_GET_AC_PARAMETERS,
    API_R1S1_GET_TOTAL_ENERGY_DATA
)
import json


class FoxR1S1Device(FoxBaseDevice):
    """F&F Fox R1S1 device. Single switch, relay with energy meter."""

    def __init__(self, device_data: DeviceData):
        """Initialize object by given device data."""
        super().__init__(device_data.name, device_data.host, device_data.api_key,
                        device_data.mac_addr, device_data.type)
        self.__device_api_client = self.DeviceRestApiImplementer(super())
        self.has_sensor_data = True
        self.total_energy_data = self.EnergySensorData
        self.ac_parameters_data = self.ACParamsSensorData

    class EnergySensorData(RestApiBaseResponse):
        """Energy sensor data holder."""

        def __init__(self, active_energy: str = None, reactive_energy: str = None, active_energy_import: str = None,
                    reactive_energy_import: str = None, status: str = API_RESPONSE_STATUS_FAIL) -> None:
            """Initialize obiect."""
            super().__init__(status)
            self.active_energy = active_energy
            self.reactive_energy = reactive_energy
            self.active_energy_import = active_energy_import
            self.reactive_energy_import = reactive_energy_import

    class ACParamsSensorData(RestApiBaseResponse):
        """AC Parameters data holder."""

        def __init__(self, voltage: str = None, current: str = None, power_active: str = None,
                    power_reactive: str = None, frequency: str = None, power_factor: str = None,
                    status: str = API_RESPONSE_STATUS_FAIL) -> None:
            """Initialize obiect."""
            super().__init__(status)
            self.voltage = voltage
            self.current = current
            self.power_active = power_active
            self.power_reactive = power_reactive
            self.frequency = frequency
            self.power_factor = power_factor

    class DeviceRestApiImplementer(RestApiClient):
        """RestAPI methods definition used by Fox R1S1 device."""

        def __init__(self, restApiClient: RestApiClient) -> None:
            """Initialize obiect."""
            self._restApiClient = restApiClient

        async def async_fetch_ac_parameters_data(self):
            """Fetch AC parametrs from device.

            Get information about electricty network such as voltage, current etc.
            see ACParamsSensorData to show what data can be readed.
            """
            device_response = await self._restApiClient.async_make_api_call_get(API_R1S1_GET_AC_PARAMETERS)
            if device_response is None:
                return FoxR1S1Device.ACParamsSensorData(status=API_RESPONSE_STATUS_FAIL)
            return FoxR1S1Device.ACParamsSensorData(**json.loads(device_response))

        async def async_fetch_total_energy_data(self):
            """Fetch total energy parametrs from device."""
            device_response = await self._restApiClient.async_make_api_call_get(API_R1S1_GET_TOTAL_ENERGY_DATA)
            if device_response is None:
                return FoxR1S1Device.EnergySensorData(status=API_RESPONSE_STATUS_FAIL)
            return FoxR1S1Device.EnergySensorData(**json.loads(device_response))

    def fetch_sensor_value_by_key(self, key: str):
        """Fetch readed value from device by key.

        If key not exist NaN str will be returned.

        Keyword arguments:
        key -- related value by key.
        Return:
        returned value by provided key.
        """
        self.all_sensor_values = {
            "voltage": self.ac_parameters_data.voltage,
            "current": self.ac_parameters_data.current,
            "power_active": self.ac_parameters_data.power_active,
            "power_reactive": self.ac_parameters_data.power_reactive,
            "frequency": self.ac_parameters_data.frequency,
            "power_factor": self.ac_parameters_data.power_factor,
            "active_energy": self.total_energy_data.active_energy,
            "reactive_energy": self.total_energy_data.reactive_energy,
            "active_energy_import": self.total_energy_data.active_energy_import,
            "reactive_energy_import": self.total_energy_data.reactive_energy_import,
        }
        try:
            return self.all_sensor_values[key]
        except KeyError:
            return "NaN"

    def get_all_electricty_data(self) -> dict:
        """Get all readed data from electricty sensor as dictionary."""
        return self.all_sensor_values

    def is_on(self, channel: int = None):
        """Return device is on status."""
        return self._state

    async def async_fetch_update(self):
        """Fetch all available data from device."""
        self._state = await self.async_fetch_channel_state()
        self.total_energy_data = await self.__device_api_client.async_fetch_total_energy_data()
        self.ac_parameters_data = await self.__device_api_client.async_fetch_ac_parameters_data()


"""Fox dimmable device implementation."""

import logging
from .fox_base_device import DeviceData, FoxBaseDevice
from foxrestapiclient.connection.rest_api_client import RestApiClient
from foxrestapiclient.connection.rest_api_responses import RestApiBaseResponse, RestApiBrightnessResponse
from foxrestapiclient.connection.const import (
    API_RESPONSE_STATUS_FAIL,
    API_RESPONSE_STATUS_INVALID,
    REQUEST_CHANNEL_KEY
)
from .const import (
    API_DIMMABLE_GET_BRIGHTNESS,
    API_DIMMABLE_SET_BRIGHTNESS,
)
import json

class FoxDimmableDevice(FoxBaseDevice):
    """Fox Dimmable device implementation. Base class for any device with dimmable feature."""

    def __init__(self, device_data: DeviceData):
        """Initialize object."""
        super().__init__(device_data.name, device_data.host, device_data. api_key,
                        device_data.mac_addr, device_data.type)
        #Extened RestApi methods, specific for device
        self.__device_api_client = self.__DeviceRestApiImplementer(super())

    class __DeviceRestApiImplementer:
        """Inner class with specific RestApi methods definition used by device."""

        def __init__(self, restApiClient: RestApiClient) -> None:
            """Initialize object."""
            self._restApiClient = restApiClient

        async def async_get_brightness_value(self, params) -> RestApiBrightnessResponse:
            """Get brightness value by given channel.

            Keyword arguments:
            params -- params dictionary, should contain channel number
            Return:
            RestApiBrightnessResponse
            """
            device_response = await self._restApiClient.async_make_api_call_get(
                API_DIMMABLE_GET_BRIGHTNESS, params)
            if device_response is None:
                return RestApiBrightnessResponse(status=API_RESPONSE_STATUS_FAIL)
            return RestApiBrightnessResponse(**json.loads(device_response))

        async def async_set_brighntess_value(self, params) -> RestApiBaseResponse:
            """Set brightness value.

            Keyword arguments:
            params -- params dictionary, should contain channel number
            Return:
            RestApiBaseResponse
            """
            device_response = await self._restApiClient.async_make_api_call_get(
                API_DIMMABLE_SET_BRIGHTNESS, params)
            if device_response is None:
                return RestApiBaseResponse(API_RESPONSE_STATUS_FAIL)
            return RestApiBaseResponse(**json.loads(device_response))

    async def async_fetch_channel_brightness(self, channel: int = 0) -> list:
        """Fetch device brightness value by given channel.

        If channel number not provided fetch brightness from all channels.

        Keyword arguments:
        channel -- channel number
        Return:
        list(int) -- readed values.
        """
        params = None
        if channel != 0:
            params = {
                REQUEST_CHANNEL_KEY: str(channel)
            }
        device_response = await self.__device_api_client.async_get_brightness_value(params)
        if (device_response.status == API_RESPONSE_STATUS_FAIL or
            device_response.status == API_RESPONSE_STATUS_INVALID):
            return [0,0]
        values = []
        if device_response.brightness != -1:
            values.append(device_response.brightness)
        if device_response.channel_1_value != -1:
            values.append(device_response.channel_1_value)
        if device_response.channel_2_value != -1:
            values.append(device_response.channel_2_value)
        return values

    async def async_update_channel_brightness(self, brightness: int, channel: int = None):
        """Set brightness value on device.

        Keyword arguments:
        brightness -- value from range <0-255>
        channel -- value from range <1-2> or None.
        """
        if brightness < 0 or brightness > 255:
            logging.warning("Brightness passed to sync_update_channel_brightness() is out of range.")
            return
        if channel is not None and (channel < 0 or channel > 2):
            logging.warning("Channel passed to sync_update_channel_brightness() is out of range.")
            return
        params = {
            "value": brightness, #Will be removed in next version.
        }
        if channel is not None:
            params.update({REQUEST_CHANNEL_KEY: channel})
        device_response = await self.__device_api_client.async_set_brighntess_value(params)
        if (device_response.status == API_RESPONSE_STATUS_FAIL or
            device_response.status == API_RESPONSE_STATUS_INVALID):
            logging.error("Setting brightness value in async_update_channel_brightness() failed.")


import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="foxrestapiclient",
    version="0.1.4",
    description="Connect to F&F Fox devices via RestAPI.",
    long_description=README,
    long_description_content_type="text/markdown",
    author="F&F Filipowski Sp. j.",
    author_email="t.waclawiak@fif.com.pl",
    license="MIT",
    url = "https://github.com/fandf92/foxrestapiclient",
    packages=find_packages(exclude=("test")),
    include_package_data=True,
    install_requires=[
        'asyncio',
        'aiohttp'
    ],
)